﻿using System;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;

namespace MonsterCSharp
{
    static class MonsterStruc
    {
        const int NoOfTraps = 2;
        const int NSDistance = 5;
        const int WEDistance = 7;

        [Serializable]
        public struct CellReference
        {
            public int NoOfCellsSouth;
            public int NoOfCellsEast;
        }

        [Serializable]
        public struct GameData
        {
            public CellReference[] TrapPositions;
            public CellReference MonsterPosition;
            public CellReference PlayerPosition;
            public CellReference FlaskPosition;
            public bool MonsterAwake;
        }

        public static void Main()
        {
            Console.WriteLine("================STRUCTURED VERSION==========");
            char[,] Cavern = new char[NSDistance+1, WEDistance+1];
            int Choice;
            CellReference FlaskPosition = new CellReference();
            bool MonsterAwake = false;
            CellReference MonsterPosition = new CellReference();
            CellReference PlayerPosition = new CellReference();
            CellReference[] TrapPositions = new CellReference[3];
            //Randomize();
            do
            {
                DisplayMenu();
                Choice = GetMainMenuChoice();
                switch (Choice)
                {
                    case 1:
                        {
                            SetUpGame(ref Cavern, ref TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake, true);
                            PlayGame(ref Cavern, TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake);
                            break;
                        }

                    case 2:
                        {
                            LoadGame(ref TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake);
                            SetUpGame(ref Cavern, ref TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake, false);
                            PlayGame(ref Cavern, TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake);
                            break;
                        }

                    case 3:
                        {
                            SaveGame(TrapPositions, MonsterPosition, PlayerPosition, FlaskPosition, MonsterAwake);
                            break;
                        }

                    case 4:
                        {
                            SetUpTrainingGame(ref Cavern, ref TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake);
                            PlayGame(ref Cavern, TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake);
                            break;
                        }
                }
            }
            while (Choice != 9);
        }

        public static void DisplayMenu()
        {
            Console.WriteLine("MAIN MENU");
            Console.WriteLine();
            Console.WriteLine("1.  Start new game");
            Console.WriteLine("2.  Load game");
            Console.WriteLine("3.  Save game");
            Console.WriteLine("4.  Play training game");
            Console.WriteLine("9.  Quit");
            Console.WriteLine();
            Console.Write("Please enter your choice: ");
        }

        public static int GetMainMenuChoice()
        {
            int Choice;
            Choice = System.Convert.ToInt32(Console.ReadLine());
            Console.WriteLine();
            return Choice;
        }

        public static void ResetCavern(ref char[,] Cavern)
        {
            int Count1;
            int Count2;
            for (Count1 = 1; Count1 <= NSDistance; Count1++)
            {
                for (Count2 = 1; Count2 <= WEDistance; Count2++)
                    Cavern[Count1, Count2] = ' ';
            }
        }

        public static CellReference GetNewRandomPosition()
        {
            CellReference Position;
            Random random = new Random();
            do
            {
                //Position.NoOfCellsSouth = (int)(VBMath.Rnd() * NSDistance) + 1;
                //Position.NoOfCellsEast = (int)(VBMath.Rnd() * WEDistance) + 1;
                Position.NoOfCellsSouth = random.Next(NSDistance) + 1;
                Position.NoOfCellsEast = random.Next(WEDistance) + 1;
            }
            while (Position.NoOfCellsSouth == 1 && Position.NoOfCellsEast == 1);
            // a random coordinate of (1,1) needs to be rejected as this is the starting position of the player
            return Position;
        }

        public static void SetPositionOfItem(ref char[,] Cavern, ref CellReference ObjectPosition, char Item, bool NewGame)
        {
            CellReference Position;
            if (NewGame && Item != '*')
            {
                do
                    Position = GetNewRandomPosition();
                while (Cavern[Position.NoOfCellsSouth, Position.NoOfCellsEast] != ' ');
                ObjectPosition = Position;
            }
            Cavern[ObjectPosition.NoOfCellsSouth, ObjectPosition.NoOfCellsEast] = Item;
        }

        public static void SetUpGame(ref char[,] Cavern, ref CellReference[] TrapPositions, ref CellReference MonsterPosition, ref CellReference PlayerPosition, ref CellReference FlaskPosition, ref bool MonsterAwake, bool NewGame)
        {
            int Count;
            ResetCavern(ref Cavern);
            if (NewGame)
            {
                PlayerPosition.NoOfCellsSouth = 1;
                PlayerPosition.NoOfCellsEast = 1;
                MonsterAwake = false;
            }
            for (Count = 1; Count <= NoOfTraps; Count++)
                SetPositionOfItem(ref Cavern, ref TrapPositions[Count], 'T', NewGame);
            SetPositionOfItem(ref Cavern, ref MonsterPosition, 'M', NewGame);
            SetPositionOfItem(ref Cavern, ref FlaskPosition, 'F', NewGame);
            SetPositionOfItem(ref Cavern, ref PlayerPosition, '*', NewGame);
        }

        public static void SetUpTrainingGame(ref char[,] Cavern, ref CellReference[] TrapPositions, ref CellReference MonsterPosition, ref CellReference PlayerPosition, ref CellReference FlaskPosition, ref bool MonsterAwake)
        {
            ResetCavern(ref Cavern);
            PlayerPosition.NoOfCellsSouth = 3;
            PlayerPosition.NoOfCellsEast = 5;
            MonsterAwake = false;
            TrapPositions[1].NoOfCellsSouth = 2;
            TrapPositions[1].NoOfCellsEast = 7;
            TrapPositions[2].NoOfCellsSouth = 4;
            TrapPositions[2].NoOfCellsEast = 5;
            MonsterPosition.NoOfCellsSouth = 1;
            MonsterPosition.NoOfCellsEast = 4;
            FlaskPosition.NoOfCellsSouth = 5;
            FlaskPosition.NoOfCellsEast = 6;
            SetUpGame(ref Cavern, ref TrapPositions, ref MonsterPosition, ref PlayerPosition, ref FlaskPosition, ref MonsterAwake, false);
        }

        public static void LoadGame(ref CellReference[] TrapPositions, ref CellReference MonsterPosition, ref CellReference PlayerPosition, ref CellReference FlaskPosition, ref bool MonsterAwake)
        {
            string Filename;
            GameData LoadedGameData = new GameData();
            Console.Write("Enter the name of the file to load: ");
            Filename = Console.ReadLine();
            Console.WriteLine();
            using (FileStream file = File.OpenRead(Filename))
            {
                BinaryFormatter reader = new BinaryFormatter();
                LoadedGameData = (GameData)reader.Deserialize(file);
            }
            TrapPositions = LoadedGameData.TrapPositions;
            MonsterPosition = LoadedGameData.MonsterPosition;
            PlayerPosition = LoadedGameData.PlayerPosition;
            FlaskPosition = LoadedGameData.FlaskPosition;
            MonsterAwake = LoadedGameData.MonsterAwake;
        }

        public static void SaveGame(CellReference[] TrapPositions, CellReference MonsterPosition, CellReference PlayerPosition, CellReference FlaskPosition, bool MonsterAwake)
        {
            string Filename;
            GameData CurrentGameData;
            CurrentGameData.TrapPositions = TrapPositions;
            CurrentGameData.MonsterPosition = MonsterPosition;
            CurrentGameData.PlayerPosition = PlayerPosition;
            CurrentGameData.FlaskPosition = FlaskPosition;
            CurrentGameData.MonsterAwake = MonsterAwake;
            Console.Write("Enter new file name: ");
            Filename = Console.ReadLine();
            Console.WriteLine();
            using (FileStream file = File.OpenWrite(Filename))
            {
                BinaryFormatter writer = new BinaryFormatter();
                writer.Serialize(file, CurrentGameData); 
            }
        }

        public static void DisplayCavern(char[,] Cavern, bool MonsterAwake)
        {
            int Count1;
            int Count2;
            for (Count1 = 1; Count1 <= NSDistance; Count1++)
            {
                Console.WriteLine(" ------------- ");
                for (Count2 = 1; Count2 <= WEDistance; Count2++)
                {
                    if (Cavern[Count1, Count2] == ' ' || Cavern[Count1, Count2] == '*' || (Cavern[Count1, Count2] == 'M' && MonsterAwake))
                        Console.Write('|' + Cavern[Count1, Count2].ToString());
                    else
                        Console.Write("| ");
                }
                Console.WriteLine("|");
            }
            Console.WriteLine(" ------------- ");
            Console.WriteLine();
        }

        public static void DisplayMoveOptions()
        {
            Console.WriteLine();
            Console.WriteLine("Enter N to move NORTH");
            Console.WriteLine("Enter E to move EAST");
            Console.WriteLine("Enter S to move SOUTH");
            Console.WriteLine("Enter W to move WEST");
            Console.WriteLine("Enter M to return to the Main Menu");
            Console.WriteLine();
        }

        public static char GetMove()
        {
            char Move;
            Move = Console.ReadLine().ToCharArray()[0];
            Console.WriteLine();
            return Move;
        }

        public static void MakeMove(ref char[,] Cavern, char Direction, ref CellReference PlayerPosition)
        {
            Cavern[PlayerPosition.NoOfCellsSouth, PlayerPosition.NoOfCellsEast] = ' ';
            switch (Direction)
            {
                case 'N':
                    {
                        PlayerPosition.NoOfCellsSouth = PlayerPosition.NoOfCellsSouth - 1;
                        break;
                    }

                case 'S':
                    {
                        PlayerPosition.NoOfCellsSouth = PlayerPosition.NoOfCellsSouth + 1;
                        break;
                    }

                case 'W':
                    {
                        PlayerPosition.NoOfCellsEast = PlayerPosition.NoOfCellsEast - 1;
                        break;
                    }

                case 'E':
                    {
                        PlayerPosition.NoOfCellsEast = PlayerPosition.NoOfCellsEast + 1;
                        break;
                    }
            }
            Cavern[PlayerPosition.NoOfCellsSouth, PlayerPosition.NoOfCellsEast] = '*';
        }

        public static bool CheckValidMove(CellReference PlayerPosition, char Direction)
        {
            bool ValidMove;
            ValidMove = true;
            if (!(Direction == 'N' || Direction == 'S' || Direction == 'W' || Direction == 'E' || Direction == 'M'))
                ValidMove = false;
            return ValidMove;
        }

        public static bool CheckIfSameCell(CellReference FirstCellPosition, CellReference SecondCellPosition)
        {
            bool InSameCell;
            InSameCell = false;
            if (FirstCellPosition.NoOfCellsSouth == SecondCellPosition.NoOfCellsSouth && FirstCellPosition.NoOfCellsEast == SecondCellPosition.NoOfCellsEast)
                InSameCell = true;
            return InSameCell;
        }

        public static void DisplayWonGameMessage()
        {
            Console.WriteLine("Well done!  You have found the flask containing the Styxian potion.");
            Console.WriteLine("You have won the game of MONSTER!");
            Console.WriteLine();
        }

        public static void DisplayTrapMessage()
        {
            Console.WriteLine("Oh no!  You have set off a trap.  Watch out, the monster is now awake!");
            Console.WriteLine();
        }

        public static void MoveFlask(ref char[,] Cavern, CellReference NewCellForFlask, ref CellReference FlaskPosition)
        {
            Cavern[NewCellForFlask.NoOfCellsSouth, NewCellForFlask.NoOfCellsEast] = 'F';
            Cavern[FlaskPosition.NoOfCellsSouth, FlaskPosition.NoOfCellsEast] = ' ';
            FlaskPosition = NewCellForFlask;
        }

        public static void MakeMonsterMove(ref char[,] Cavern, ref CellReference MonsterPosition, ref CellReference FlaskPosition, CellReference PlayerPosition)
        {
            CellReference OriginalMonsterPosition;
            bool MonsterMovedToSameCellAsFlask;
            OriginalMonsterPosition = MonsterPosition;
            Cavern[MonsterPosition.NoOfCellsSouth, MonsterPosition.NoOfCellsEast] = ' ';
            if (MonsterPosition.NoOfCellsSouth < PlayerPosition.NoOfCellsSouth)
                MonsterPosition.NoOfCellsSouth = MonsterPosition.NoOfCellsSouth + 1;
            else if (MonsterPosition.NoOfCellsSouth > PlayerPosition.NoOfCellsSouth)
                MonsterPosition.NoOfCellsSouth = MonsterPosition.NoOfCellsSouth - 1;
            else if (MonsterPosition.NoOfCellsEast < PlayerPosition.NoOfCellsEast)
                MonsterPosition.NoOfCellsEast = MonsterPosition.NoOfCellsEast + 1;
            else
                MonsterPosition.NoOfCellsEast = MonsterPosition.NoOfCellsEast - 1;
            MonsterMovedToSameCellAsFlask = CheckIfSameCell(MonsterPosition, FlaskPosition);
            if (MonsterMovedToSameCellAsFlask)
                MoveFlask(ref Cavern, OriginalMonsterPosition, ref FlaskPosition);
            Cavern[MonsterPosition.NoOfCellsSouth, MonsterPosition.NoOfCellsEast] = 'M';
        }

        public static void DisplayLostGameMessage()
        {
            Console.WriteLine("ARGHHHHHH!  The monster has eaten you.  GAME OVER.");
            Console.WriteLine("Maybe you will have better luck next time you play MONSTER!");
            Console.WriteLine();
        }

        public static void PlayGame(ref char[,] Cavern, CellReference[] TrapPositions, ref CellReference MonsterPosition, ref CellReference PlayerPosition, ref CellReference FlaskPosition, ref bool MonsterAwake)
        {
            int Count;
            bool Eaten;
            bool FlaskFound;
            char MoveDirection;
            bool ValidMove;
            Eaten = false;
            FlaskFound = false;
            DisplayCavern(Cavern, MonsterAwake);
            do
            {
                do
                {
                    DisplayMoveOptions();
                    MoveDirection = GetMove();
                    ValidMove = CheckValidMove(PlayerPosition, MoveDirection);
                }
                while (!ValidMove);
                if (MoveDirection != 'M')
                {
                    MakeMove(ref Cavern, MoveDirection, ref PlayerPosition);
                    DisplayCavern(Cavern, MonsterAwake);
                    FlaskFound = CheckIfSameCell(PlayerPosition, FlaskPosition);
                    if (FlaskFound)
                        DisplayWonGameMessage();
                    Eaten = CheckIfSameCell(MonsterPosition, PlayerPosition);
                    if (!MonsterAwake && !FlaskFound && !Eaten)
                    {
                        MonsterAwake = CheckIfSameCell(PlayerPosition, TrapPositions[1]);
                        if (!MonsterAwake)
                            MonsterAwake = CheckIfSameCell(PlayerPosition, TrapPositions[2]);
                        if (MonsterAwake)
                        {
                            DisplayTrapMessage();
                            DisplayCavern(Cavern, MonsterAwake);
                        }
                    }

                    if (MonsterAwake && !Eaten && !FlaskFound)
                    {
                        Count = 0;
                        do
                        {
                            MakeMonsterMove(ref Cavern, ref MonsterPosition, ref FlaskPosition, PlayerPosition);
                            Eaten = CheckIfSameCell(MonsterPosition, PlayerPosition);
                            Console.WriteLine();
                            Console.WriteLine("Press Enter key to continue");
                            Console.ReadLine();
                            DisplayCavern(Cavern, MonsterAwake);
                            Count = Count + 1;
                        }
                        while (Count != 2 && !Eaten);
                    }

                    if (Eaten)
                        DisplayLostGameMessage();
                }
            }
            while (!Eaten && !FlaskFound && MoveDirection != 'M');
        }
    }
}
